;********************************************************************************************************
;                                                                     					*
;    Programma di test per decoder NMRA con decodifica ad interrupt   					*
;                                                                     					*
;    Autore: DF - (www.seven-segments.com)                            					*
;                                                                     					*
;	Il programma esegue un controllo del segnale dcc ad intervalli					*
;	di 44 us: questo e' il valore piu' alto per avere 2 o 3 conteggi 				*
;	su un bit a "1" (periodo min 104us e max 128us) e almeno 4 					*
;	conteggi su un bit a "0" (periodo min 180us e max 12000us).Il 					*
;	funzionamento e' critico su processori a 4MHz e tranquillo su					*
;	processori a 8MHz. Analisi dell'occupazione del tempo cpu:					*
;													*
;	1. tempo impiegato quando il segnale dcc e' a 0 o senza fronte in salita: 	25 cicli	*
;	2. tempo impiegato quando si riceve un "1" in attesa preambolo:			36 cicli	*
;	3. tempo impiegato quando c'e' un preambolo valido o meno:			42 cicli	*
;	4. tempo impiegato quando si memorizza un dato:					40 cicli	*
;	5. tempo di controllo del bit di start/stop:					40/41 cicli	*
;													*
;	quindi considerando un pacchetto di 5ms ogni 5ms (circa 100 pacchetti/sec)			*
;	trasmissione di bit a "1" tra un pacchetto e l'altro (caso peggiore) si avranno			*
;	circa 3/4 dei bit a "1" e 1/4 dei bit a "0" (mediamente) quindi ogni secondo			*
;	si trasmetteranno 6466 bit "1" e 1078 bit a "0", quindi (considerando che 			*
;	l'interrupt viene chiamato 22727 volte al secondo):						*
;													*
;	Ttot = (22727/2)*25 + (6466+1078)*42 + (22727/2-6466-1078)*36 = 				*
;	     = 11364*25+7544*42+3820*36 = 738468							*
;	     												*
;	su un totale di 1000000 di istruzioni a 4MHz e 20000000 a 8MHz (unici				*
;	casi per avere PIC senza quarzo esterno). Penso che a 4MHz si debba usare			*
;	anche la NMRA standard S9.2 riga 118, interrompendo per 5 ms la ricezione			*
;	dei pacchetti dopo l'arrivo di un pacchetto indirizzato al decoder....				*
;													*
;													*
;********************************************************************************************************		



        list      p=12f675                                      ; definizione del processore
        include <p12f675.inc>                                   ; variabili caratteristiche del processore

        __CONFIG _CP_OFF & _CPD_OFF & _BODEN_OFF & _MCLRE_ON & _PWRTE_ON & _WDT_ON & _INTRC_OSC_NOCLKOUT



;**********************************************************************
;                                                                     *
;    Definizione costanti                                             *
;                                                                     *
;**********************************************************************

#DEFINE         dccin           GPIO,0x5                         ;input del segnale dcc
#DEFINE         forward         GPIO,0x1                         ;bit che controlla la marcia avanti
#DEFINE         backward        GPIO,0x2                         ;bit che controlla la marcia indietro
#DEFINE         out1            GPIO,0x0                         ;output funzione 1
#DEFINE         out2            GPIO,0x4                         ;output funzione 2

#DEFINE         olddcc          flag,0x0                         ;stato dell'input dcc all'ultimo campionamento
#DEFINE         newpack         flag,0x1                         ;nuovo pacchetto dcc disponibile

;**********************************************************************
;                                                                     *
;    Common Variable definition                                       *
;                                                                     *
;**********************************************************************

_w              EQU   0x20                      ; salvataggio del registro w durante l'interrupt
_status         EQU   0x21                      ; salvataggio del registro status durante l'interrupt
_pclath         EQU   0x22                      ; salvataggio del registro pclath durante l'interrupt
dcc_count       EQU   0x23                      ; conteggio del periodo del segnale dcc
flag            EQU   0x24                      ; flag[7..0]
						; flag[0] olddcc
						; flag[0] newpack
state           EQU   0x25                      ; stato di decodifica del segnale 
						; 0 - attesa di un preambolo valido
						; 1 - caricamento di un dato generico
						; 2 - controllo del prossimo start/stop bit
						
one_count       EQU   0x26                      ; Conta il numero di '1' nel preambolo
bit_count       EQU   0x27                      ; Conta il numero di bit per byte (8)
pwm             EQU   0x29                      ; valore pwm 
ioport          EQU   0x2A                      ; replica la porta di uscita GPIO
pwmc            EQU   0x2B                      ; contatore/accumulatore pwm


buffer_start    EQU   0x50                      ; inizio del buffer per la ricezione di pacchetti dcc
buffer1         EQU   0x51
buffer2         EQU   0x52
buffer3         EQU   0x53
buffer4         EQU   0x54
buffer5         EQU   0x55
buffer6         EQU   0x56
buffer7         EQU   0x57                      ; fine del buffer di ricezione di pacchetti dcc


;**********************************************************************
;                                                                     *
;    Inizio programma                                                 *
;                                                                     *
;**********************************************************************

                ORG     0x000                   	; indirizzo di start del processore 

                clrf    PCLATH                  	; pagina iniziale (sul 675/629 non dovrebbe servire)
                goto    main                            ; vai all'inizio del programma


;**********************************************************************
;                                                                     *
;       Interrupt routine                                             *
;                                                                     *
;**********************************************************************

                ORG     0x004                           ; indirizzo di interrupt del processore

Interrupt       movwf   _w                              ;1      salva w
                swapf   STATUS,w                        ;2      scambia w e status
                clrf    STATUS                          ;3      cancella i flag di stato, punta ai registri di pagina 0 (sul 675/629 non dovrebbe servire)
                movwf   _status                         ;4      salva status

                bcf     INTCON,T0IF                     ;5      e' abilitato solo l'interrupt da timer0, cancella la richiesta
                movlw   d'256'- d'44'+ d'10'            ;6      un interrupt ogni 44us
                movwf   TMR0                            ;7      inizia il conteggio

                btfss   dcc_count,0x2                   ;8      se il contatore di periodo non e' arrivato a 4...
                incf    dcc_count,f                     ;9      ...incrementalo

                movf    pwm,w                           ;10	controllo pwm: carica il valore
                addwf   pwmc,f                          ;11	sommalo nell'accumulatore
                btfss   STATUS,C                        ;12	se c'e' overflow setta l'uscita...
                goto    drive_off                       ;13-14  ...altrimenti resettala

drive_on        movf    ioport,w                        ;14	in ioport ci deve essere la configurazione per marcia avanti o indietro
                movwf   GPIO                            ;15	...mettila in uscita
                btfss   dccin                           ;16	campiona il segnale dcc
                goto    endint_bit0                     ;17 -> 25 se e' 0 allora esci subito...
                btfsc   olddcc                          ;18     se il segnale dcc e' 1 ma il vecchio campione era 1...
                goto    endint                          ;19 -> 26 ...esci
                bsf     olddcc                          ;20     se arriviamo qui c'e' un fronte in salita...
                movf    state,w                         ;21     a che punto della decodifica siamo?
                addwf   PCL,F                           ;22     salta nello stato giusto
                goto    sync_check                      ;23-24  attesa del preambolo
                goto    data_load                       ;23-24  memorizzazione di un byte generico (address/data/xor...)
                goto    bit_check                       ;23-24  controllo di fine del pacchetto
                
drive_off       bcf     forward                         ;15	qui siamo nella parte "off" del periodo pwm
                bcf     backward                        ;16	resetta le uscite al ponte di pilotaggio motore
                btfss   dccin                           ;17	stati uguali a quelli sopra...
                goto    endint_bit0                     ;18 -> 26
                btfsc   olddcc                          ;19     
                goto    endint                          ;20 -> 27 
                bsf     olddcc                          ;21     
                movf    state,w                         ;22     
                addwf   PCL,F                           ;23     
                goto    sync_check                      ;24-25  
                goto    data_load                       ;24-25  
                goto    bit_check                       ;24-25  

endint_bit0     bcf     olddcc                          ;	setta a "0" il bit di campionamento dcc
endint  	swapf   _status,W                       ;	
                movwf   STATUS                          ;      	ritorna al vecchio status
                swapf   _w,F                            ;
                swapf   _w,W                            ;      	ritorna la vecchio w
                retfie                                  ;	uscita dall'interrupt

                
;**********************************************************************
;                                                                     *
;       Attesa del preambolo                                          *
;                                                                     *
;**********************************************************************
                
sync_check      btfsc   dcc_count,0x2                   ;26     se il dcc_count e' minore di 4 il bit e' un "1"...
                goto    bit_zero                        ;27-28  ...altrimenti e' uno "0"

bit_one 	btfss   one_count,0x4                   ;28     se il conteggio di "1" e' inferiore a 16...
                incf    one_count,f                     ;29     ...incrementa il numero di "1" ricevuti
                clrf    dcc_count                       ;30     azzera il contatore di periodo dcc
                swapf   _status,W                       ;31
                movwf   STATUS                          ;32
                swapf   _w,F                            ;33
                swapf   _w,W                            ;34
                retfie                                  ;35-36 	uscita dall'interrupt (attesa dello start bit)

bit_zero        movlw   0xF6                            ;29-30  questo e' uno start bit se il numero
                addwf   one_count,w                     ;31     di "1" ricevuti e' superiore a 10..
                btfss   STATUS,C                        ;32     ...preambolo ok
                goto    clrexit                         ;33-34 uscita dall'interrupt per preambolo non valido
                incf    state,f                         ;34     vai nello stato di memorizzazione del dato
                clrf    one_count                       ;35     azzera il contatore di "1" ricevuti...
                clrf    dcc_count                       ;36     ...ed il contatore per il periodo dcc
                swapf   _status,W                       ;37
                movwf   STATUS                          ;38
                swapf   _w,F                            ;39
                swapf   _w,W                            ;40
                retfie                                  ;41-42 uscita dall'interrupt per preambolo ok

clrexit     	clrf    dcc_count                       ;35    azzera il contatore del periodo dcc...
                clrf    one_count                       ;36    ... ed il contatore di "1"
                swapf   _status,W                       ;37
                movwf   STATUS                          ;38
                swapf   _w,F                            ;39
                swapf   _w,W                            ;40    
                retfie                                  ;41-42 esci dall'interrupt (attesa di un preambolo valido)


;**********************************************************************
;                                                                     *
;       Caricamento del prossimo dato dcc                             *
;                                                                     *
;**********************************************************************
                
data_load       rlf     INDF,f                          ;26     ruota a sinistra il dato
                btfsc   dcc_count,0x2                   ;27     le prossime 4 istruzioni in effetti copiano
                bcf     INDF,0x0                        ;28     il bit ricevuto nel buffer di ingresso dei dati...
                btfss   dcc_count,0x2                   ;29
                bsf     INDF,0x0                        ;30     
                clrf    dcc_count                       ;31     azzera il contatore di periodo
                decf    bit_count,f                     ;32     decrementa il contatore di bit
                btfsc   STATUS,Z                        ;33     se il contatore e' 0, allora ho ricevuto un byte
                incf    state,f                         ;34     vai allo stato di controllo dello start/stop bit
                swapf   _status,W                       ;35
                movwf   STATUS                          ;36
                swapf   _w,F                            ;37
                swapf   _w,W                            ;38     
                retfie                                  ;39-40	esci dall'interrupt


;**********************************************************************
;                                                                     *
;       Controllo del prossimo start/stop bit                         *
;                                                                     *
;**********************************************************************
                
bit_check       btfsc   dcc_count,0x2                   ;26	se il dcc_count e' minore di 4 il bit e' un "1"...
                goto    bbit_zero                       ;27-28	...altrimenti e' uno zero

bbit_one        bsf     newpack                         ;28     e' arrivato un nuovo pacchetto!
                clrf    state                           ;29     riparti ad aspettare un preambolo
                movlw   buffer_start                    ;30     reinizializza... 
                movwf   FSR                             ;31     il puntatore all'inizio del buffer di ricezione
                movlw   0x8                             ;32     
                movwf   bit_count                       ;33     reinizializza il contatore di bit...
                clrf    dcc_count                       ;34     ed il contatore del periodo dcc
                swapf   _status,W                       ;35
                movwf   STATUS                          ;36
                swapf   _w,F                            ;37
                swapf   _w,W                            ;38    
                retfie                                  ;39-40 uscita dall'interrupt (nuovo pacchetto ricevuto...)

bbit_zero       incf    FSR,f                           ;29     il pacchetto non e' finito, punta al prossimo byte
                btfsc   FSR,0x3                         ;30	se l'indirizzo e' oltre la fine buffer...
                goto    bbit_one                        ;31     ...tronca la ricezione del pacchetto perche' c'e' un errore
                decf    state,f                         ;32     altrimenti torna al caricamento del prossimo byte
                movlw   0x8                             ;33
                movwf   bit_count                       ;34     reinizializza il contatore dei bit
                clrf    dcc_count                       ;35     reinizializza il contatore del periodo
                swapf   _status,W                       ;36
                movwf   STATUS                          ;37
                swapf   _w,F                            ;38
                swapf   _w,W                            ;39     
                retfie                                  ;40-41 esci dall'interrupt per pacchetto incompleto


;**********************************************************************
;                                                                     *
;    Inizio del programma                                             *
;                                                                     *
;**********************************************************************

main            banksel GPIO
                clrf    GPIO
                banksel TRISIO
                movlw   b'00101000'
                movwf   TRISIO                         
                banksel CMCON
                movlw 07h 
                movwf CMCON 
                banksel ANSEL
                clrf ANSEL				;inizializzazioni di base...

                banksel 0x0
                clrf    dcc_count
                clrf    flag     
                clrf    state    
                clrf    one_count
                clrf    ioport
                clrf    pwm
                clrf    pwmc
                movlw   0x8
                movwf   bit_count
                movlw   buffer_start                    
                movwf   FSR      			;inizializzazione delle variabili

                banksel OSCCAL
                movlw   0xa8 				; qui ci va messo il valore di calibrazione o un piu' tradizionale                                   
                movwf   OSCCAL                          ; call 0x3FF

                banksel OPTION_REG
                movlw   b'11011111'                                     
                movwf   OPTION_REG                      ;abilita timer0 e disabilita i pull-up
                banksel INTCON
                movlw   b'10100000'                                     
                movwf   INTCON                          ;interrupt solo dal timer0, abilitati

forever 	clrwdt					;programma di test...
                btfss   newpack                         ;controllo se ho ricevuto un nuovo pacchetto...
                goto    forever				; altrimenti non faccio nulla...
                bcf     newpack                         ;resetto il pacchetto ricevuto
                movlw   0x3				
                subwf   buffer_start,w
                btfss   STATUS,Z                        ; se l'indirizzo primario e' "3"...
                goto    forever
                
                movlw   b'11100000'
                andwf   buffer1,w
                sublw   b'01100000'             
                btfss   STATUS,Z                        ; ...se ho ricevuto un comando di "marcia avanti" (011)...
                goto    forever
                
                movf    buffer_start,w
                xorwf   buffer1,w
                subwf   buffer2,w
                btfss   STATUS,Z                        ; e il checksum e' ok...
                goto    forever
                
                rlf     buffer1,f
                rlf     buffer1,f
                rlf     buffer1,w
                andlw   b'11111000'
                movwf   pwm                             ; allora metto il valore nel registro del pwm...
		btfss	forward				; ...ed imposto la marcia avanti...
		btfsc	backward
		
                goto    forever


;**********************************************************************
;                                                                     *
;       Fine del codice                                               *
;                                                                     *
;**********************************************************************

                end
